package info.frostfs.sdk.pool;

import info.frostfs.sdk.enums.HealthyStatus;
import info.frostfs.sdk.enums.MethodIndex;
import info.frostfs.sdk.utils.FrostFSMessages;
import lombok.Getter;
import lombok.Setter;
import org.slf4j.Logger;

import java.util.Arrays;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;

@Getter
@Setter
public class ClientStatusMonitor implements ClientStatus {
    private final ReentrantLock lock = new ReentrantLock();
    private final Logger logger;
    private final AtomicInteger healthy = new AtomicInteger();

    private final String address;
    private final MethodStatus[] methods;

    private int errorThreshold;
    private int currentErrorCount;
    private long overallErrorCount;

    public ClientStatusMonitor(Logger logger, String address) {
        this.logger = logger;
        this.healthy.set(HealthyStatus.HEALTHY.value);

        this.address = address;
        this.methods = Arrays.stream(MethodIndex.values())
                .map(t -> new MethodStatus(t.methodName))
                .toArray(MethodStatus[]::new);
    }

    @Override
    public boolean isHealthy() {
        return healthy.get() == HealthyStatus.HEALTHY.value;
    }

    @Override
    public boolean isDialed() {
        return healthy.get() != HealthyStatus.UNHEALTHY_ON_DIAL.value;
    }

    public void setHealthy() {
        healthy.set(HealthyStatus.HEALTHY.ordinal());
    }

    @Override
    public void setUnhealthy() {
        healthy.set(HealthyStatus.UNHEALTHY_ON_REQUEST.value);
    }

    public void setUnhealthyOnDial() {
        healthy.set(HealthyStatus.UNHEALTHY_ON_DIAL.value);
    }

    public void incErrorRate() {
        boolean thresholdReached;
        lock.lock();
        try {
            currentErrorCount++;
            overallErrorCount++;

            thresholdReached = currentErrorCount >= errorThreshold;

            if (thresholdReached) {
                setUnhealthy();
                currentErrorCount = 0;
            }
        } finally {
            lock.unlock();
        }

        if (thresholdReached && logger != null) {
            FrostFSMessages.errorThresholdReached(logger, address, errorThreshold);
        }
    }

    @Override
    public int getCurrentErrorRate() {
        lock.lock();
        try {
            return currentErrorCount;
        } finally {
            lock.unlock();
        }
    }

    @Override
    public long getOverallErrorRate() {
        lock.lock();
        try {
            return overallErrorCount;
        } finally {
            lock.unlock();
        }
    }

    @Override
    public StatusSnapshot[] getMethodsStatus() {
        StatusSnapshot[] result = new StatusSnapshot[methods.length];

        for (int i = 0; i < result.length; i++) {
            result[i] = methods[i].getSnapshot();
        }

        return result;
    }
}

