package info.frostfs.sdk.services.impl;

import frostfs.container.ContainerServiceGrpc;
import frostfs.container.Service;
import frostfs.refs.Types;
import info.frostfs.sdk.dto.container.Container;
import info.frostfs.sdk.dto.container.ContainerId;
import info.frostfs.sdk.dto.session.SessionToken;
import info.frostfs.sdk.enums.StatusCode;
import info.frostfs.sdk.enums.WaitExpects;
import info.frostfs.sdk.exceptions.ResponseFrostFSException;
import info.frostfs.sdk.exceptions.TimeoutFrostFSException;
import info.frostfs.sdk.jdo.ClientEnvironment;
import info.frostfs.sdk.jdo.parameters.CallContext;
import info.frostfs.sdk.jdo.parameters.PrmWait;
import info.frostfs.sdk.jdo.parameters.container.PrmContainerCreate;
import info.frostfs.sdk.jdo.parameters.container.PrmContainerDelete;
import info.frostfs.sdk.jdo.parameters.container.PrmContainerGet;
import info.frostfs.sdk.jdo.parameters.container.PrmContainerGetAll;
import info.frostfs.sdk.jdo.parameters.session.SessionContext;
import info.frostfs.sdk.mappers.container.ContainerIdMapper;
import info.frostfs.sdk.mappers.container.ContainerMapper;
import info.frostfs.sdk.mappers.netmap.VersionMapper;
import info.frostfs.sdk.mappers.object.OwnerIdMapper;
import info.frostfs.sdk.services.ContainerClient;
import info.frostfs.sdk.services.ContextAccessor;
import info.frostfs.sdk.tools.RequestConstructor;
import info.frostfs.sdk.tools.RequestSigner;
import info.frostfs.sdk.tools.Verifier;
import info.frostfs.sdk.utils.WaitUtil;

import java.time.LocalDateTime;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static info.frostfs.sdk.constants.AttributeConst.DISABLE_HOMOMORPHIC_HASHING_ATTRIBUTE;
import static info.frostfs.sdk.utils.DeadLineUtil.deadLineAfter;
import static info.frostfs.sdk.utils.Validator.validate;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;

public class ContainerClientImpl extends ContextAccessor implements ContainerClient {
    private final ContainerServiceGrpc.ContainerServiceBlockingStub serviceBlockingStub;
    private final SessionToolsImpl sessionTools;

    public ContainerClientImpl(ClientEnvironment clientEnvironment) {
        super(clientEnvironment);
        this.serviceBlockingStub = ContainerServiceGrpc.newBlockingStub(clientEnvironment.getChannel());
        this.sessionTools = new SessionToolsImpl(clientEnvironment);
    }

    public SessionToken getOrCreateSession(SessionContext sessionContext, CallContext ctx) {
        return isNull(sessionContext.getSessionToken())
                ? sessionTools.getOrCreateSession(getContext(), ctx)
                : sessionContext.getSessionToken();
    }

    @Override
    public Container getContainer(PrmContainerGet args, CallContext ctx) {
        validate(args);

        var request = createGetRequest(args);

        var service = deadLineAfter(serviceBlockingStub, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.get(request);

        Verifier.checkResponse(response);
        return ContainerMapper.toModel(response.getBody().getContainer());
    }

    @Override
    public List<ContainerId> listContainers(PrmContainerGetAll args, CallContext ctx) {
        validate(args);

        var request = createListRequest(args);

        var service = deadLineAfter(serviceBlockingStub, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.list(request);

        Verifier.checkResponse(response);

        return response.getBody().getContainerIdsList().stream()
                .map(cid -> new ContainerId(cid.getValue().toByteArray()))
                .collect(Collectors.toList());
    }

    @Override
    public ContainerId createContainer(PrmContainerCreate args, CallContext ctx) {
        validate(args);

        var request = createPutRequest(args, ctx);

        var service = deadLineAfter(serviceBlockingStub, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.put(request);

        Verifier.checkResponse(response);

        waitForContainer(WaitExpects.EXISTS, response.getBody().getContainerId(), args.getWaitParams());

        return new ContainerId(response.getBody().getContainerId().getValue().toByteArray());
    }

    @Override
    public void deleteContainer(PrmContainerDelete args, CallContext ctx) {
        validate(args);

        var request = createDeleteRequest(args, ctx);

        var service = deadLineAfter(serviceBlockingStub, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.delete(request);

        Verifier.checkResponse(response);

        waitForContainer(WaitExpects.REMOVED, request.getBody().getContainerId(), args.getWaitParams());
    }

    private void waitForContainer(WaitExpects expect, Types.ContainerID cid, PrmWait waitParams) {
        var request = createGetRequest(cid, null);

        waitParams = isNull(waitParams) ? new PrmWait() : waitParams;
        var deadLine = waitParams.getDeadline();

        while (true) {
            try {
                var response = serviceBlockingStub.get(request);
                Verifier.checkResponse(response);

                if (expect == WaitExpects.EXISTS) {
                    break;
                }

                if (LocalDateTime.now().isAfter(deadLine)) {
                    throw new TimeoutFrostFSException();
                }

                WaitUtil.sleep(waitParams.getPollInterval().toMillis());
            } catch (ResponseFrostFSException exp) {
                if (LocalDateTime.now().isAfter(deadLine)) {
                    throw new TimeoutFrostFSException();
                }

                if (exp.getStatus().getCode() != StatusCode.CONTAINER_NOT_FOUND) {
                    throw exp;
                }

                if (expect == WaitExpects.REMOVED) {
                    break;
                }

                WaitUtil.sleep(waitParams.getPollInterval().toMillis());
            }
        }
    }

    private Service.GetRequest createGetRequest(PrmContainerGet args) {
        var cid = ContainerIdMapper.toGrpcMessage(args.getContainerId());
        return createGetRequest(cid, args.getXHeaders());
    }

    private Service.GetRequest createGetRequest(Types.ContainerID cid, Map<String, String> xHeaders) {
        var body = Service.GetRequest.Body.newBuilder()
                .setContainerId(cid)
                .build();
        var request = Service.GetRequest.newBuilder()
                .setBody(body);

        RequestConstructor.addMetaHeader(request, xHeaders);
        RequestSigner.sign(request, getContext().getKey());

        return request.build();
    }

    private Service.ListRequest createListRequest(PrmContainerGetAll args) {
        var body = Service.ListRequest.Body.newBuilder()
                .setOwnerId(OwnerIdMapper.toGrpcMessage(getContext().getOwnerId()))
                .build();
        var request = Service.ListRequest.newBuilder()
                .setBody(body);

        RequestConstructor.addMetaHeader(request, args.getXHeaders());
        RequestSigner.sign(request, getContext().getKey());

        return request.build();
    }

    private Service.PutRequest createPutRequest(PrmContainerCreate args, CallContext ctx) {
        syncContainerWithNetwork(args.getContainer(), ctx);
        var container = ContainerMapper.toGrpcMessage(args.getContainer()).toBuilder()
                .setOwnerId(OwnerIdMapper.toGrpcMessage(getContext().getOwnerId()))
                .setVersion(VersionMapper.toGrpcMessage(getContext().getVersion()))
                .build();

        var body = Service.PutRequest.Body.newBuilder()
                .setContainer(container)
                .setSignature(RequestSigner.signRFC6979(getContext().getKey(), container))
                .build();
        var request = Service.PutRequest.newBuilder()
                .setBody(body);

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createContainerTokenContext(
                sessionToken,
                null,
                frostfs.session.Types.ContainerSessionContext.Verb.PUT,
                container.getOwnerId(),
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        RequestSigner.sign(request, getContext().getKey());

        return request.build();
    }

    private Service.DeleteRequest createDeleteRequest(PrmContainerDelete args, CallContext ctx) {
        var cid = ContainerIdMapper.toGrpcMessage(args.getContainerId());

        var body = Service.DeleteRequest.Body.newBuilder()
                .setContainerId(cid)
                .setSignature(RequestSigner.signRFC6979(getContext().getKey(), cid.getValue()))
                .build();
        var request = Service.DeleteRequest.newBuilder()
                .setBody(body);

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createContainerTokenContext(
                sessionToken,
                null,
                frostfs.session.Types.ContainerSessionContext.Verb.DELETE,
                null,
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        RequestSigner.sign(request, getContext().getKey());

        return request.build();
    }

    private void syncContainerWithNetwork(Container container, CallContext callContext) {
        var settings = getContext().getFrostFSClient().getNetworkSettings(callContext);
        if (nonNull(settings.getHomomorphicHashingDisabled()) && settings.getHomomorphicHashingDisabled()) {
            container.getAttributes().put(DISABLE_HOMOMORPHIC_HASHING_ATTRIBUTE, Boolean.TRUE.toString());
        } else {
            container.getAttributes().remove(DISABLE_HOMOMORPHIC_HASHING_ATTRIBUTE, Boolean.TRUE.toString());
        }
    }
}
