package info.frostfs.sdk.tools;

import com.google.protobuf.ByteString;
import com.google.protobuf.Message;
import frostfs.session.Types;
import info.frostfs.sdk.dto.response.MetaHeader;
import info.frostfs.sdk.dto.session.SessionToken;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import info.frostfs.sdk.jdo.ECDsa;
import info.frostfs.sdk.mappers.response.MetaHeaderMapper;
import info.frostfs.sdk.mappers.session.SessionMapper;
import org.apache.commons.collections4.MapUtils;

import java.util.Map;
import java.util.stream.Collectors;

import static info.frostfs.sdk.constants.ErrorConst.REQUIRED_PROTO_FIELD_TEMPLATE;
import static info.frostfs.sdk.constants.FieldConst.META_HEADER_FIELD_NAME;
import static info.frostfs.sdk.tools.RequestSigner.signMessagePart;
import static info.frostfs.sdk.utils.MessageHelper.getField;
import static info.frostfs.sdk.utils.MessageHelper.setField;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;

public class RequestConstructor {
    private RequestConstructor() {
    }

    public static void addMetaHeader(Message.Builder request) {
        addMetaHeader(request, null, null);
    }

    public static void addMetaHeader(Message.Builder request, Map<String, String> xHeaders) {
        addMetaHeader(request, xHeaders, null);
    }

    public static void addMetaHeader(Message.Builder request,
                                     Map<String, String> xHeaders,
                                     Types.SessionToken sessionToken) {
        if (isNull(request)) {
            return;
        }

        if (isNull(request.getDescriptorForType().findFieldByName(META_HEADER_FIELD_NAME))) {
            throw new ValidationFrostFSException(String.format(REQUIRED_PROTO_FIELD_TEMPLATE, META_HEADER_FIELD_NAME));
        }

        if (((Types.RequestMetaHeader) getField(request, META_HEADER_FIELD_NAME)).getSerializedSize() > 0) {
            return;
        }

        var metaHeader = MetaHeaderMapper.toGrpcMessageBuilder(new MetaHeader());

        if (nonNull(sessionToken) && sessionToken.getSerializedSize() > 0) {
            metaHeader.setSessionToken(sessionToken);
        }

        if (MapUtils.isNotEmpty(xHeaders)) {
            var grpcXHeaders = xHeaders.entrySet()
                    .stream()
                    .map(e -> Types.XHeader.newBuilder().setKey(e.getKey()).setValue(e.getValue()).build())
                    .collect(Collectors.toList());

            metaHeader.addAllXHeaders(grpcXHeaders);
        }

        setField(request, META_HEADER_FIELD_NAME, metaHeader.build());
    }

    public static Types.SessionToken createObjectTokenContext(SessionToken sessionToken,
                                                              frostfs.refs.Types.Address address,
                                                              Types.ObjectSessionContext.Verb verb,
                                                              ECDsa key) {
        var protoToken = SessionMapper.deserializeSessionToken(sessionToken.getToken());

        var target = Types.ObjectSessionContext.Target.newBuilder()
                .setContainer(address.getContainerId());

        if (address.getObjectId().getSerializedSize() > 0) {
            target.addObjects(address.getObjectId());
        }

        var ctx = Types.ObjectSessionContext.newBuilder()
                .setTarget(target.build())
                .setVerb(verb)
                .build();
        var body = protoToken.getBody().toBuilder()
                .setObject(ctx)
                .setSessionKey(ByteString.copyFrom(key.getPublicKeyByte()))
                .build();

        return protoToken.toBuilder()
                .setSignature(signMessagePart(key, body))
                .setBody(body)
                .build();
    }

    public static Types.SessionToken createContainerTokenContext(SessionToken sessionToken,
                                                                 frostfs.refs.Types.ContainerID containerId,
                                                                 Types.ContainerSessionContext.Verb verb,
                                                                 frostfs.refs.Types.OwnerID ownerId,
                                                                 ECDsa key) {
        var protoToken = SessionMapper.deserializeSessionToken(sessionToken.getToken());

        var containerSession = Types.ContainerSessionContext.newBuilder().setVerb(verb);

        if (isNull(containerId) || containerId.getSerializedSize() == 0) {
            containerSession.setWildcard(true);
        } else {
            containerSession.setContainerId(containerId);
        }

        var bodyBuilder = protoToken.getBody().toBuilder()
                .setContainer(containerSession)
                .setSessionKey(ByteString.copyFrom(key.getPublicKeyByte()));

        if (nonNull(ownerId) && ownerId.getSerializedSize() > 0) {
            bodyBuilder.setOwnerId(ownerId);
        }

        var body = bodyBuilder.build();

        return protoToken.toBuilder()
                .setSignature(signMessagePart(key, body))
                .setBody(body)
                .build();
    }
}
