package info.frostfs.sdk.mappers.container;

import com.google.protobuf.ByteString;
import frostfs.container.Types;
import info.frostfs.sdk.dto.container.Container;
import info.frostfs.sdk.enums.BasicAcl;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import info.frostfs.sdk.mappers.netmap.PlacementPolicyMapper;
import info.frostfs.sdk.mappers.netmap.VersionMapper;
import org.apache.commons.collections4.CollectionUtils;

import java.util.stream.Collectors;

import static info.frostfs.sdk.UuidExtension.asBytes;
import static info.frostfs.sdk.UuidExtension.asUuid;
import static info.frostfs.sdk.constants.ErrorConst.UNKNOWN_ENUM_VALUE_TEMPLATE;
import static java.util.Objects.isNull;

public class ContainerMapper {
    private ContainerMapper() {
    }

    public static Types.Container toGrpcMessage(Container container) {
        if (isNull(container)) {
            return null;
        }

        var containerGrpc = Types.Container.newBuilder()
                .setBasicAcl(container.getBasicAcl().value)
                .setPlacementPolicy(PlacementPolicyMapper.toGrpcMessage(container.getPlacementPolicy()))
                .setNonce(ByteString.copyFrom(asBytes(container.getNonce())));

        var attributes = container.getAttributes().entrySet().stream()
                .map(entry ->
                        Types.Container.Attribute.newBuilder()
                                .setKey(entry.getKey())
                                .setValue(entry.getValue())
                                .build()
                )
                .collect(Collectors.toList());
        containerGrpc.addAllAttributes(attributes);

        return containerGrpc.build();
    }

    public static Container toModel(Types.Container containerGrpc) {
        if (isNull(containerGrpc) || containerGrpc.getSerializedSize() == 0) {
            return null;
        }

        var basicAcl = BasicAcl.get(containerGrpc.getBasicAcl());
        if (isNull(basicAcl)) {
            throw new ProcessFrostFSException(
                    String.format(UNKNOWN_ENUM_VALUE_TEMPLATE, BasicAcl.class.getName(), containerGrpc.getBasicAcl())
            );
        }


        var container = new Container(basicAcl, PlacementPolicyMapper.toModel(containerGrpc.getPlacementPolicy()));
        container.setNonce(asUuid(containerGrpc.getNonce().toByteArray()));
        container.setVersion(VersionMapper.toModel(containerGrpc.getVersion()));

        if (CollectionUtils.isNotEmpty(containerGrpc.getAttributesList())) {
            var attributes = containerGrpc.getAttributesList().stream()
                    .collect(Collectors.toMap(Types.Container.Attribute::getKey, Types.Container.Attribute::getValue));
            container.setAttributes(attributes);
        }

        return container;
    }
}
