package info.frostfs.sdk.mappers.object;

import com.google.protobuf.ByteString;
import frostfs.object.Types;
import info.frostfs.sdk.dto.container.ContainerId;
import info.frostfs.sdk.dto.netmap.Version;
import info.frostfs.sdk.dto.object.ObjectAttribute;
import info.frostfs.sdk.dto.object.ObjectHeader;
import info.frostfs.sdk.dto.object.OwnerId;
import info.frostfs.sdk.enums.ObjectType;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.MockedStatic;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mockStatic;

public class ObjectHeaderMapperTest {

    @ParameterizedTest
    @EnumSource(value = ObjectType.class)
    void toGrpcMessage_success(ObjectType type) {
        //Given
        var containerId = new ContainerId("EQGx2QeYHJb53uRwYGzcQaW191sZpdNrjutk6veUSV2R");
        var objectHeader = new ObjectHeader(
                containerId,
                type,
                new ObjectAttribute("Filename", "cat.jpg"),
                new ObjectAttribute("Filename2", "cat2.jpg")
        );
        objectHeader.setOwnerId(new OwnerId("NVxUSpEEJzYXZZtUs18PrJTD9QZkLLNQ8S"));
        objectHeader.setVersion(new Version(1, 2));

        //When
        var result = ObjectHeaderMapper.toGrpcMessage(objectHeader);

        //Then
        assertNotNull(result);
        assertEquals(
                objectHeader.getContainerId().getValue(),
                new ContainerId(result.getContainerId().getValue().toByteArray()).getValue()
        );
        assertEquals(objectHeader.getObjectType().value, result.getObjectType().getNumber());
        assertEquals(objectHeader.getAttributes().size(), result.getAttributesCount());

        assertEquals(objectHeader.getAttributes().get(0).getKey(), result.getAttributes(0).getKey());
        assertEquals(objectHeader.getAttributes().get(0).getValue(), result.getAttributes(0).getValue());

        assertEquals(objectHeader.getAttributes().get(1).getKey(), result.getAttributes(1).getKey());
        assertEquals(objectHeader.getAttributes().get(1).getValue(), result.getAttributes(1).getValue());
    }

    @Test
    void toGrpcMessage_null() {
        //When + Then
        assertThrows(ValidationFrostFSException.class, () -> ObjectHeaderMapper.toGrpcMessage(null));
    }

    @Test
    void toGrpcMessage_notValidScheme() {
        //Given
        var containerId = new ContainerId("EQGx2QeYHJb53uRwYGzcQaW191sZpdNrjutk6veUSV2R");
        var objectHeader = new ObjectHeader(containerId, ObjectType.REGULAR);

        //When + Then
        try (MockedStatic<Types.ObjectType> mockStatic = mockStatic(Types.ObjectType.class)) {
            mockStatic.when(() -> Types.ObjectType.forNumber(objectHeader.getObjectType().value))
                    .thenReturn(null);

            assertThrows(ProcessFrostFSException.class, () -> ObjectHeaderMapper.toGrpcMessage(objectHeader));
        }
    }

    @ParameterizedTest
    @EnumSource(value = Types.ObjectType.class, names = "UNRECOGNIZED", mode = EnumSource.Mode.EXCLUDE)
    void toModel_success(Types.ObjectType type) {
        //Given
        var containerId = new ContainerId("EQGx2QeYHJb53uRwYGzcQaW191sZpdNrjutk6veUSV2R");
        var cid = frostfs.refs.Types.ContainerID.newBuilder()
                .setValue(ByteString.copyFrom(containerId.toHash()))
                .build();
        var version = frostfs.refs.Types.Version.newBuilder()
                .setMajor(1)
                .setMinor(2)
                .build();
        var header = Types.Header.newBuilder()
                .setObjectType(type)
                .setVersion(version)
                .setContainerId(cid)
                .setPayloadLength(10)
                .build();

        //When
        var result = ObjectHeaderMapper.toModel(header);

        //Then
        assertNotNull(result);
        assertThat(result.getAttributes()).hasSize(0);
        assertEquals(header.getObjectTypeValue(), result.getObjectType().value);
        assertEquals(version.getMajor(), result.getVersion().getMajor());
        assertEquals(version.getMinor(), result.getVersion().getMinor());
        assertEquals(header.getPayloadLength(), result.getPayloadLength());
        assertEquals(containerId.getValue(), result.getContainerId().getValue());
    }

    @Test
    void toModel_null() {
        //When + Then
        assertNull(ObjectHeaderMapper.toModel(null));
        assertNull(ObjectHeaderMapper.toModel(Types.Header.getDefaultInstance()));
    }

    @Test
    void toModel_notValidScheme() {
        //Given
        var header = Types.Header.newBuilder()
                .setObjectType(Types.ObjectType.TOMBSTONE)
                .build();

        //When + Then
        try (MockedStatic<ObjectType> mockStatic = mockStatic(ObjectType.class)) {
            mockStatic.when(() -> ObjectType.get(Types.ObjectType.TOMBSTONE.getNumber()))
                    .thenReturn(null);

            assertThrows(ProcessFrostFSException.class, () -> ObjectHeaderMapper.toModel(header));
        }
    }
}
